;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "glsl-mode" "20260603.1348"
  "Major mode for Open GLSL shader files."
  '((emacs "26.1"))
  :url "https://github.com/jimhourihan/glsl-mode"
  :commit "515a2ba4dab3ec89c83a962902a123ddf81e3cfe"
  :revdesc "515a2ba4dab3"
  :keywords '("languages" "opengl" "gpu" "spir-v" "vulkan")
  :authors '(("Gustaf Waldemarson" . "gustaf.waldemarson~at~gmail.com")
             ("Jim Hourihan" . "jimhourihan~at~gmail.com")
             (nil . "Xavier.Decoret@imag.fr"))
  :maintainers '(("Gustaf Waldemarson" . "gustaf.waldemarson~at~gmail.com")
                 ("Jim Hourihan" . "jimhourihan~at~gmail.com")
                 (nil . "Xavier.Decoret@imag.fr")))
